import concurrent.futures
import re
from functools import cached_property
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile

from ..base import MaskDataset
from ..types import BundledPath


class BBBC020(MaskDataset):
    """Murine bone-marrow derived macrophages

    The image set consists of 25 images, each consisting of three channels. The
    samples were stained with DAPI and CD11b/APC. In addition to this, a merged
    image is provided. DAPI labels the nuclei and CD11b/APC the cell surface.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'equal', 'cv2', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    anno_ch : {'nuclei', 'cells'}, default: ('nuclei',)
        Which channel(s) to load as annotation. Make sure to give it as a
        Sequence when choose a single channel.
    drop_missing_pairs : bool, default: True
        Valid only if `output='both'`. It will drop images that do not have mask
        pairs.

    Warnings
    --------
    5 annotations are missing: ind={17,18,19,20,21}
    [jw-30min 1, jw-30min 2, jw-30min 3, jw-30min 4, jw-30min 5]

        - ./BBBC020_v1_images/jw-30min 1/jw-30min 1_(c1+c5).TIF
        - ./BBBC020_v1_images/jw-30min 2/jw-30min 2_(c1+c5).TIF
        - ./BBBC020_v1_images/jw-30min 3/jw-30min 3_(c1+c5).TIF
        - ./BBBC020_v1_images/jw-30min 4/jw-30min 4_(c1+c5).TIF
        - ./BBBC020_v1_images/jw-30min 5/jw-30min 5_(c1+c5).TIF

    - BBC020_v1_outlines_nuclei/jw-15min 5_c5_43.TIF exists but corrupted

    Notes
    -----
    - Anotations are instance segmented where each of them is saved as a single
      image file. It loads and aggregates them as a single array. Later label
      will overlap the former ones. If you do not want this behavior, modify
      subclass this class and override ``get_mask()`` method.
    - 2 channels; R channel is the same as G, R==G!=B
        Assign 0 to red channel
    - BBBC has received a complaint that "BBB020_v1_outlines_nuclei" appears
      incomplete and we have been unable to obtain the missing images from the
      original contributor.
    - Nuclei anno looks good
    - Should separte nuclei and cells annotation; if ``anno_ch=None``,
      ``anno_dict`` becomes a mess.

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC020

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """

    # Dataset's acronym
    acronym = 'BBBC020'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        anno_ch: Sequence[str] = ('nuclei',),
        drop_missing_pairs: bool = True,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this dataset
        self._num_channels = 2  # explicit for `grayscale`
        self.anno_ch = anno_ch
        if not any([ch in ('nuclei', 'cells') for ch in anno_ch]):
            raise ValueError("Set `anno_ch` in ('nuclei', 'cells') in sequence")
        self.drop_missing_pairs = drop_missing_pairs

        if self.output == 'both' and self.drop_missing_pairs:
            self.file_list, self.anno_dict = self._drop_missing_pairs()

    def get_image(self, p: Path) -> np.ndarray:
        img = tifffile.imread(p)
        # R==G, zero 0
        img[..., 0] = 0
        return img

    def get_mask(self, lst_p: Union[BundledPath, List[BundledPath]]) -> np.ndarray:
        def _assign_index(
            mask: np.ndarray,
            fn: Union[str, Path],
            ind: int
        ):
            """For threading"""
            tif: np.ndarray = tifffile.imread(fn)
            idx_nz = tif.nonzero()
            mask[idx_nz] = ind

        if len(self.anno_ch) == 1:
            tif: np.ndarray = tifffile.imread(lst_p[0])
            mask = np.zeros_like(tif)
            idx_nz = tif.nonzero()
            mask[idx_nz] = 1
            with concurrent.futures.ThreadPoolExecutor() as executor:
                [executor.submit(_assign_index, mask, p, ind)
                 for ind, p in enumerate(lst_p[1:], 2)]
        elif len(self.anno_ch) == 2:
            # mask_cells
            lst_p_cells = lst_p[0]
            tif: np.ndarray = tifffile.imread(lst_p_cells[0])
            mask_cells = np.zeros_like(tif)
            idx_nz = tif.nonzero()
            mask_cells[idx_nz] = 1
            # mask_nuclei
            lst_p_nuclei = lst_p[1]
            tif: np.ndarray = tifffile.imread(lst_p_nuclei[0])
            mask_nuclei = np.zeros_like(tif)
            idx_nz = tif.nonzero()
            mask_nuclei[idx_nz] = 1
            # threading
            with concurrent.futures.ThreadPoolExecutor() as executor:
                [executor.submit(_assign_index, mask_cells, p, ind)
                 for ind, p in enumerate(lst_p_cells[1:], 2)]
                [executor.submit(_assign_index, mask_nuclei, p, ind)
                 for ind, p in enumerate(lst_p_nuclei[1:], 2)]
            # 'cells' (ch=Green) first then 'nuclei' (ch=Blue)
            mask = np.stack((mask_cells, mask_nuclei), axis=-1)
        return mask

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        parent = 'BBBC020_v1_images'
        file_list = sorted(root_dir.glob(f'{parent}/*/*_(c1+c5).TIF'))
        return file_list

    @staticmethod
    def _sort_key(p: Path):
        res = re.search(r'\d+$', p.stem)
        if res is None:
            raise ValueError
        return int(res.group())

    @cached_property
    def anno_dict(self) -> Dict[int, BundledPath]:
        def _filter_valid_file(p: Path):
            return p.stat().st_size > 0

        root_dir = self.root_dir
        anno_dict = {}
        for i, p in enumerate(self.file_list):
            k = p.parent.stem
            # parent = root_dir / 'BBBC020_v1_outlines_*'
            if len(ch := self.anno_ch) == 1:
                parent = root_dir / f'BBBC020_v1_outlines_{ch[0]}'
                anno_list = sorted(
                    parent.glob(f'{k}_*.TIF'),
                    key=self._sort_key
                )
                anno_list = list(filter(_filter_valid_file, anno_list))
            elif len(ch) == 2:
                anno_list_cells = sorted(
                    root_dir.glob(f'BBBC020_v1_outlines_nuclei/{k}_*.TIF'),
                    key=self._sort_key
                )
                anno_list_cells = list(filter(_filter_valid_file,
                                              anno_list_cells))
                anno_list_nuclei = sorted(
                    root_dir.glob(f'BBBC020_v1_outlines_cells/{k}_*.TIF'),
                    key=self._sort_key
                )
                anno_list_nuclei = list(filter(_filter_valid_file,
                                               anno_list_nuclei))
                if anno_list_cells and anno_list_nuclei:
                    # 'cells' (ch=Green) first then 'nuclei' (ch=Blue)
                    anno_list = [anno_list_cells, anno_list_nuclei]
                else:
                    anno_list = []
            else:
                raise ValueError("Set `anno_ch` in ('nuclei', 'cells')")
            if anno_list:
                anno_dict[i] = anno_list
        return anno_dict



"""Custom transforms for bioimages based on albumentations

"""

from typing import Optional, Tuple, Union

import albumentations
import cv2
import numpy as np


class HWCToCHW(albumentations.ImageOnlyTransform):
    """Transpose axes from (H, W, C) to (C, H, W)

    By default, ``bioimageloader`` returns images in shape of (H, W, C=3)
    regardless of its color mode for easy handling. Some models expect (C, H, W)
    shape of images as input. It converts (C, H, W) to (C, H, W).

    See Also
    --------
    albumentations.ImageOnlyTransform : super class

    Examples
    --------
    >>> import albumentations as A
    >>> from bioimageloader import Config

    >>> cfg = Config('config.yml')
    >>> transforms = A.Compose([
            HWCToCHW(),
        ])
    >>> datasets = cfg.load_datasets(transforms=transforms)
    >>> dset = datasets[0]  # select only the first dataset
    >>> data = dset[0]  # select only the first image
    >>> print(data['image'].shape)
    (3, H, W)

    """
    def __init__(self, always_apply: bool = False, p: float = 1.0):
        super().__init__(always_apply=always_apply, p=p)

    def apply(self, img, **params):
        # img to have (H, W, C)
        return img.transpose(-1, 0, 1)


class SqueezeGrayImageCHW(albumentations.ImageOnlyTransform):
    """Squeeze grayscale image from (3, H, W) to (1, H, W)|(H, W)

    By default, ``bioimageloader`` returns images in 3 channels regardless of
    their color mode for easy handling. If a model requires (C=1, H, W) shape of
    input, first use ``HWCToCHW`` and use this transform to convert from (3, H,
    W) to (1, H, W).

    Parameters
    ----------
    keep_dim : bool, default: True
        Keep channel axis to 1
    always_apply : bool, default: False
    p : float, default: 1.0
        Value between [0.0, 1.0]

    See Also
    --------
    albumentations.ImageOnlyTransform : super class
    bioimageloader.transforms.HWCToCHW
    bioimageloader.transforms.SqueezeGrayImageHWC

    Examples
    --------
    >>> import albumentations as A
    >>> from bioimageloader import Config

    >>> cfg = Config('config.yml')
    >>> transforms = A.Compose([
            HWCToCHW(),
            SqueezeGrayImageCHW(),
        ])
    >>> datasets = cfg.load_datasets(transforms=transforms)
    >>> dset = datasets[0]  # select only the first dataset
    >>> data = dset[0]  # select only the first image
    >>> print(data['image'].shape)
    (1, H, W)

    You can set ``keep_dim`` False to entirely drop channel axis, as some models
    require. But use ``SqueezeGrayImageHWC`` instead for that.

    >>> transforms = A.Compose([
            HWCToCHW(),
            SqueezeGrayImageCHW(keep_dim=False),  # drop channel axis
        ])
    >>> datasets = cfg.load_datasets(transforms=transforms)
    >>> dset = datasets[0]  # select only the first dataset
    >>> data = dset[0]  # select only the first image
    >>> print(data['image'].shape)
    (H, W)

    """
    def __init__(
        self,
        keep_dim=True,
        always_apply: bool = False,
        p: float = 1.0,
    ):
        super().__init__(always_apply=always_apply, p=p)
        self.keep_dim = keep_dim

    def apply(self, img, **params):
        if self.keep_dim:
            # img to have (1, H, W)
            return img[0:1]
        # img to have (H, W)
        return img[0]

    def get_transform_init_args_names(self):
        return (
            'keep_dim',
        )


class SqueezeGrayImageHWC(albumentations.ImageOnlyTransform):
    """Squeeze grayscale image from (H, W, 3) to (H, W)|(H, W, 1)

    By default, ``bioimageloader`` returns images in 3 channels regardless of
    their color mode for easy handling. It converts (H, W, 3) to (H, W).

    If a model requires (H, W) shape of input, use this transform to convert
    from (H, W, 3) to (H, W).

    Parameters
    ----------
    keep_dim : bool, default: False
    always_apply : bool, default: False
    p : float, default: 1.0
        Value between [0.0, 1.0]

    See Also
    --------
    albumentations.ImageOnlyTransform : super class
    bioimageloader.transforms.SqueezeGrayImageHWC

    Examples
    --------
    >>> import albumentations as A
    >>> from bioimageloader import Config

    >>> cfg = Config('config.yml')
    >>> transforms = A.Compose([
            SqueezeGrayImageHWC(),
        ])
    >>> datasets = cfg.load_datasets(transforms=transforms)
    >>> dset = datasets[0]  # select only the first dataset
    >>> data = dset[0]  # select only the first image
    >>> print(data['image'].shape)
    (H, W)

    """
    def __init__(
        self,
        keep_dim=False,
        always_apply: bool = False,
        p: float = 1.0,
    ):
        super().__init__(always_apply=always_apply, p=p)
        self.keep_dim = keep_dim

    def apply(self, img, **params):
        if self.keep_dim:
            # img to have (H, W, 1)
            return img[..., 0:1]
        # img to have (H, W)
        return img[..., 0]

    def get_transform_init_args_names(self):
        return (
            'keep_dim',
        )


class ExpandToRGB(albumentations.DualTransform):
    """Make sure image has 3 channels, RGB

    Expand axis of image that has 2 channels to have 3 channels mainly for
    visualization. Albumentations ver.

    Warnings
    --------
    This will be deprecated. Grayscale conversion is done in eash Dataset

    """
    def __init__(
        self,
        always_apply: bool = False,
        p: float = 1.0,
        select_mask_channel: Optional[int] = None,
    ):
        super().__init__(always_apply=always_apply, p=p)
        self.select_mask_channel = select_mask_channel

    def apply(self, img, **params):
        num_channels = img.shape[-1]
        if num_channels != 2:
            raise ValueError
        stacked = np.concatenate(
            [img, np.zeros_like(img[..., 0])[..., np.newaxis]],
            axis=-1,
        )
        # if dtype is not None:
        #     return stacked.astype(dtype)
        return stacked

    def apply_to_mask(self, img, **params):
        if (ch := self.select_mask_channel) is not None:
            return img[..., ch]
        return img

    def get_transform_init_args_names(self):
        return (
            'select_mask_channel',
        )


class RGBToGray(albumentations.ImageOnlyTransform):
    """ToGray preserve all 3 channels from the input. This transform truncates
    channels dimension.

    Warnings
    --------
    This will be deprecated. Grayscale conversion is done in eash Dataset

    Notes
    -----
    Targets:
        image

    Image types:
        uint8, float32
    """

    def __init__(self, always_apply: bool = False, p: float = 1.0):
        super().__init__(always_apply=always_apply, p=p)

    def apply(self, img, **params):
        return cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

    def get_transform_init_args_names(self):
        return ()


class ToGrayBySum(albumentations.ImageOnlyTransform):
    """Convert image to gray scale by tacking mean of existing channels

    For 2 channels, multi-modal images, ToGray does not make sense. Normally,
    rgb2gray conversions is a linear sum of RGB values (opencv [1]_, pillow
    [2]_). Just summing with eqaul weights would be more correct for bioimages.

    Warnings
    --------
    This will be deprecated. Grayscale conversion is done through ``grayscale``
    and ``grayscale_mode`` arguments in eash Dataset

    References
    ----------
    .. [1] https://docs.opencv.org/3.4/de/d25/imgproc_color_conversions.html#color_convert_rgb_gray
    .. [2] https://pillow.readthedocs.io/en/stable/reference/Image.html#PIL.Image.Image.convert

    """
    def __init__(
            self,
            always_apply: bool = False,
            p: float = 1.0,
            num_channels: Optional[int] = None,
    ):
        super().__init__(always_apply=always_apply, p=p)
        self.num_channels = num_channels

    def apply(self, img, **params):
        dtype = img.dtype
        if self.num_channels is not None:
            img = np.sum(img, axis=-1) / self.num_channels
            return img.astype(dtype)
        return np.mean(img, axis=-1).astype(dtype)


class ChannelReorder(albumentations.ImageOnlyTransform):
    """Reorder channel

    Expect images with 3 channels. Reorder and make it continuous in 'C' order.

    Parameters
    ----------
    order : tuple of three integers
        Reorder by indexing
    always_apply : bool, default: False
    p : float, default: 1.0
        Value between [0.0, 1.0]

    See Also
    --------
    albumentations.ImageOnlyTransform : super class
    albumentations.augmentations.transforms.ChannelShuffle : random shuffling

    Examples
    --------
    >>> import numpy as np
    >>> from bioimageloader.transforms import ChannelReorder

    >>> arr = np.arange(12).reshape((2, 2, 3))
    >>> print(arr)
    [[[ 0  1  2]
      [ 3  4  5]]
     [[ 6  7  8]
      [ 9 10 11]]]

    >>> reorder = ChannelReorder((2, 1, 0))
    >>> arr_reordered = reorder.apply(arr)
    >>> print(arr_reordered)
    [[[ 2  1  0]
      [ 5  4  3]]
     [[ 8  7  6]
      [11 10  9]]]

    """
    def __init__(
        self,
        order: Tuple[int, int, int],
        always_apply: bool = False,
        p: float = 1.0,
    ):
        super().__init__(always_apply=always_apply, p=p)
        self.order = order

    def apply(self, img, **params):
        return np.ascontiguousarray(img[..., self.order])

    def get_transform_init_args_names(self):
        return (
            'order',
        )


class NormalizePercentile(albumentations.ImageOnlyTransform):
    """Normalize using percentile

    Compute q-th percentile min- and max-values from given image array and
    normalize.

    Use ``numpy.percentile()`` [1]_

    Expect images with 3 channels. Reorder and make it continuous in 'C' order.

    Parameters
    ----------
    qmin : float, default: 0.0
        Lower bound quantile in range of [0, 100)
    qmax : float, default: 99.8
        Upper bound quantile in range of (0, 100]
    per_channel : bool, default: False
        Whether to calculate percentile per channel or not
    clip : bool, default: False
        Whether to clip in [0, 1] or not. Read more in Returns section.
    order : tuple of three integers
        Reorder by indexing
    always_apply : bool, default: False
    p : float, default: 1.0
        Value between [0.0, 1.0]

    Returns
    -------
    img_norm
        Normalized image in float32 in range of [0.0, 1.0] if ``clip`` set to
        True, else its value overflows lower beyond 0.0 and higher beyond 1.0.

    See Also
    --------
    albumentations.ImageOnlyTransform : super class
    albumentations.augmentations.transforms.ChannelShuffle : random shuffling

    References
    ----------
    .. [1] https://numpy.org/doc/stable/reference/generated/numpy.percentile.html

    Examples
    --------
    >>> import numpy as np
    >>> from bioimageloader.transforms import ChannelReorder

    >>> arr = np.arange(12).reshape((2, 2, 3))
    >>> print(arr)
    [[[ 0  1  2]
      [ 3  4  5]]
     [[ 6  7  8]
      [ 9 10 11]]]

    >>> reorder = ChannelReorder((2, 1, 0))
    >>> arr_reordered = reorder.apply(arr)
    >>> print(arr_reordered)
    [[[ 2  1  0]
      [ 5  4  3]]
     [[ 8  7  6]
      [11 10  9]]]

    """
    def __init__(
        self,
        qmin: float = 0.0,
        qmax: float = 99.8,
        per_channel: bool = False,
        clip: bool = False,
        always_apply: bool = False,
        p: float = 1.0,
    ):
        super().__init__(always_apply=always_apply, p=p)
        self.qmin = qmin
        self.qmax = qmax
        self.per_channel = per_channel
        self._per_channel = (0, 1) if per_channel else None
        self.clip = clip

    def apply(self, img, **params):
        if self.qmin == 0:
            vmin = 0.0
            vmax = np.percentile(img, self.qmax,
                                 axis=self._per_channel, keepdims=True)
        else:
            v = np.percentile(img, (self.qmin, self.qmax),
                              axis=self._per_channel, keepdims=True)
            vmin = v[0]
            vmax = v[1]
        img = (img - vmin) / (vmax - vmin)
        if self.clip:
            img = np.clip(img, 0, 1)
        return img.astype(np.float32)

    def get_transform_init_args_names(self):
        return (
            'qmin',
            'qmax',
            'per_channel',
            'clip',
        )



from functools import cached_property
from pathlib import Path
from typing import List, Optional, Sequence, Union

import albumentations
import numpy as np

from ..base import Dataset
from ..types import BundledPath
from ..utils import bundle_list, imread_asarray, stack_channels_to_rgb


class BBBC015(Dataset):
    """Human U2OS cells transfluor

    The images are of a human osteosarcoma cell line (U2OS) co-expressing beta2
    (b2AR) adrenergic receptor and arrestin-GFP protein molecules. The receptor
    was modified-type that generates "vesicle-type" spots upon ligand
    stimulation.

    The plate was acquired on iCyte imaging cytometer with iCyte software
    version 2.5.1. Image file format is JPEG with one image for green channel
    and one image for crimson channel. Image size is 1000 x 768 pixels.

    This image set has a portion of a 96-well plate containing 3 replica rows
    and 12 concentration points of isoproterenol. In each well four fields were
    acquired. File name structure: <well-number>_<field>_<channel>.JPG

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    image_ch : {'b2AR', 'arrestin'}, default: ('b2AR', 'arrestin')
        Which channel(s) to load as image. Make sure to give it as a Sequence
        when choose a single channel.

    Notes
    -----
    - 2 channels (Green, Crimson?), texture in green channel
    - Crimson channel...?
    - RGB channel is all the same in each image file

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC015

    See Also
    --------
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC015'

    def __init__(
        self,
        root_dir: str,
        *,
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        image_ch: Sequence[str] = ('b2AR', 'arrestin'),
        **kwargs
    ):
        self._root_dir = root_dir
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this dataset
        self.image_ch = image_ch
        if not any([ch in ('b2AR', 'arrestin') for ch in image_ch]):
            raise ValueError("Set `image_ch` in ('b2AR', 'arrestin') in sequence")

    def get_image(self, p: Union[Path, BundledPath]) -> np.ndarray:
        # RGB is all the same
        if isinstance(p, Path):
            img = imread_asarray(p)
        else:
            img = stack_channels_to_rgb(
                lambda x: imread_asarray(x)[..., 0], p, 1, 0, 2
            )
        return img

    @cached_property
    def file_list(self) -> Union[List[Path], List[BundledPath]]:
        root_dir = self.root_dir
        parent = 'BBBC015_v1_images'
        file_list = sorted(root_dir.glob(f'{parent}/*.JPG'))
        if len(ch := self.image_ch) == 1:
            if ch[0] == 'b2AR':
                return file_list[::2]
            elif ch[0] == 'arrestin':
                return file_list[1::2]
        return bundle_list(file_list, 2)



"""Collection of public bioimage datasets
"""

# import should not be sorted by isort
# #--- MaskDataset (mask anno) ---# #
#   full anno
#     instance
from ._dsb2018 import DSB2018
from ._compath import ComputationalPathology
from ._frunet import FRUNet
from ._s_bsst265 import S_BSST265
from ._bbbc006 import BBBC006
from ._bbbc020 import BBBC020
from ._bbbc039 import BBBC039
#     semantic (fg/bg)
from ._tnbc import TNBC
from ._bbbc008 import BBBC008
#     semantic (boundary, outline)
from ._murphylab import MurphyLab
from ._bbbc007 import BBBC007
from ._bbbc018 import BBBC018
#   partial anno
#     instance
#     semantic (fg/bg)
from ._digitpath import DigitalPathology
from ._ucsb import UCSB
# #--- Dataset (no anno) ---# #
from ._bbbc002 import BBBC002  # a few annotated
from ._bbbc013 import BBBC013
from ._bbbc014 import BBBC014
from ._bbbc015 import BBBC015
from ._bbbc016 import BBBC016
from ._bbbc026 import BBBC026
from ._bbbc041 import BBBC041
from ._bbbc021 import BBBC021  # huge dataset 132,000 images


# Keep this list sorted
__all__ = [
    'BBBC002',
    'BBBC006',
    'BBBC007',
    'BBBC008',
    'BBBC013',
    'BBBC014',
    'BBBC015',
    'BBBC016',
    'BBBC018',
    'BBBC020',
    'BBBC021',
    'BBBC026',
    'BBBC039',
    'BBBC041',
    'ComputationalPathology',
    'DSB2018',
    'DigitalPathology',
    'FRUNet',
    'MurphyLab',
    'S_BSST265',
    'TNBC',
    'UCSB',
]

assert __all__ == sorted(__all__), "Keep collections sorted"



from functools import cached_property
from pathlib import Path
from typing import List, Optional

import albumentations
import cv2
import numpy as np
import tifffile

from ..base import Dataset


class BBBC002(Dataset):
    """Drosophila Kc167 cells

    There are 10 fields of view of each sample, for a total of 50 fields of
    view. The images were acquired on a Zeiss Axiovert 200M microscope. The
    images provided here are a single channel, DNA. The image size is 512 x 512
    pixels. The images are provided as 8-bit TIFF files.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.

    Notes
    -----
    - Cell count available
    - ImageJ RoI available for 3 tiles
        - CPvalid1_48_40x_Tiles_p0151DAPI_ROIs.zip
        - CPvalid1_340_40x_Tiles_p1175DAPI_ROIs.zip
        - CPvalid1_nodsRNA_40x_Tiles_p0219DAPI_ROIs.zip

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC002

    See Also
    --------
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC002'

    def __init__(
        self,
        root_dir: str,
        *,
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        **kwargs
    ):
        self._root_dir = root_dir
        self._transforms = transforms
        self._num_calls = num_calls

    def get_image(self, p: Path) -> np.ndarray:
        img = tifffile.imread(p)
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        return img

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        parent = 'drosophila_kc167_1_images'
        file_list = sorted(root_dir.glob(f'{parent}/*.TIF'))
        return file_list



from functools import cached_property, partial
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import albumentations
import numpy as np
import tifffile
from PIL import Image

from ..base import MaskDataset
from ..types import BundledPath
from ..utils import bundle_list, stack_channels, stack_channels_to_rgb


class BBBC018(MaskDataset):
    """Human HT29 colon-cancer cells (diverse phenotypes)

    The image set consists of 56 fields of view (4 from each of 14 samples).
    Because there are three channels, there are 168 image files. (The samples
    were stained with Hoechst 33342, pH3, and phalloidin. Hoechst 33342 is a DNA
    stain that labels the nucleus. Phospho-histone H3 indicates mitosis.
    Phalloidin labels actin, which is present in the cytoplasm.) The samples are
    the top-scoring sample from each of Jones et al.'s classifiers, as listed in
    the file SamplesScores.zip in their supplement. The files are in DIB format,
    as produced by the Cellomics ArrayScan instrument at the Whitehead–MIT
    Bioimaging Center. We recommend using Bio-Formats to read the DIB files.
    Each image is 512 x 512 pixels.

    The filenames are of the form wellidx-channel.DIB, where wellidx is the
    five-digit well index (from Jones et al.'s supplement) and channel is either
    DNA, actin, or pH3, depending on the channel.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'equal', 'cv2', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    anno_ch : {'DNA', 'actin'}, default: ('DNA',)
        Which channel(s) to load as annotation. Make sure to give it as a
        Sequence when choose a single channel.
    drop_missing_pairs : bool, default: True
        Valid only if `output='both'`. It will drop images that do not have mask
        pairs.

    Other Parameters
    ----------------
    image_ch : {'DNA', 'actin'}, default: ('DNA', 'actin')
        Which channel(s) to load as image. Make sure to give it as a Sequence
        when choose a single channel.

    Warnings
    --------
    BBBC018_v1_images/10779 annotation is missing. len(anno_dict) =
    len(file_list) - 1; ind={26}

        - PosixPath('images/bbbc/018/BBBC018_v1_images/10779-DNA.DIB')
        - PosixPath('images/bbbc/018/BBBC018_v1_images/10779-actin.DIB')
        - PosixPath('images/bbbc/018/BBBC018_v1_images/10779-pH3.DIB')

    Notes
    -----
    - Every DIB has 3 channels (Order = (DNA,actin,pH3)). The second one is the
      object.
    - DNA -> Nuceli
    - Actin -> Cell
    - Annotation is outline one, but every anno is closed so binary_fill_holes
      works fine
    - For some reason annotation is y inverted

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC018

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """

    # Dataset's acronym
    acronym = 'BBBC018'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        # image_ch: Sequence[str] = ('DNA', 'actin',),
        anno_ch: Sequence[str] = ('DNA',),
        drop_missing_pairs: bool = True,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_calls = num_calls
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # self.image_ch = image_ch
        self.anno_ch = anno_ch
        self.drop_missing_pairs = drop_missing_pairs

        if self.output == 'both' and self.drop_missing_pairs:
            self.file_list, self.anno_dict = self._drop_missing_pairs()

    @staticmethod
    def _imread_handler(p: Path) -> np.ndarray:
        img = Image.open(p)
        return np.asarray(img)[..., 1]

    def get_image(self, p: BundledPath) -> np.ndarray:
        # Second channel has objects
        # Order = (DNA,actin,pH3)
        img = stack_channels_to_rgb(self._imread_handler, p, 2, 0, 1)
        return img

    def get_mask(self, p: Union[Path, BundledPath]) -> np.ndarray:
        if isinstance(p, Path):
            mask = np.asarray(Image.open(p))
        else:
            mask = stack_channels(Image.open, p)
        if mask.dtype == 'bool':
            mask = 255 * mask.astype(np.uint8)
        # For some reason mask is -y
        return np.ascontiguousarray(mask[::-1, ...])

    @cached_property
    def file_list(self) -> List[BundledPath]:
        root_dir = self.root_dir
        parent = 'BBBC018_v1_images'
        # Order = (DNA,actin,pH3)
        _file_list = sorted(root_dir.glob(f'{parent}/*.DIB'))
        return bundle_list(_file_list, 3)

    @cached_property
    def anno_dict(self) -> Union[Dict[int, Path], Dict[int, BundledPath]]:
        root_dir = self.root_dir
        parent = 'BBBC018_v1_outlines'
        # _anno_list = sorted(root_dir.glob(f'{parent}/*.png'))
        stain_to_target = {'DNA': 'nuclei',
                           'actin': 'cells'}
        if len(ch := self.anno_ch) == 1:
            anno_dict: Dict[int, Path] = {}
            target = stain_to_target[ch[0]]
            for i, p in enumerate(self.file_list):
                name = p[0].stem.split('-')[0]
                fn = root_dir / parent / f'{name}-{target}.png'
                if fn.exists():
                    anno_dict[i] = fn
            return anno_dict
        elif len(ch) == 2:
            anno_bdict: Dict[int, BundledPath] = {}
            for i, p in enumerate(self.file_list):
                name = p[0].stem.split('-')[0]
                lst_fn = []
                for c in ch:
                    target = stain_to_target[c]
                    fn = root_dir / parent / f'{name}-{target}.png'
                    if fn.exists():
                        lst_fn.append(fn)
                if lst_fn:
                    anno_bdict[i] = lst_fn
        else:
            raise ValueError
        return anno_bdict



from pathlib import Path
from typing import Optional, Sequence, Union

import albumentations
import numpy as np
from PIL import Image

from bioimageloader.collections import DSB2018


class DSB2018NUnet(DSB2018):
    """Data Science Bowl 2018 (Adopted for NU-Net)

    Each entry is a pair of an image and a mask.
    By default, it applies `base_transform`, which makes an image Tensor and
    have data range of uint8 [0, 255].

    Returns a dictionary, whose key is determined by `output` argument.
    """

    # Set acronym
    acronym = 'DSB2018'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # specific to this dataset
        training: bool = True,
        # NU-Net args
        idx_invt: Optional[Sequence[int]] = None,
        **kwargs
    ):
        """
        Parameters
        ----------
        root_dir : str
            Path to root directory
        output : {'image', 'mask', 'both'} (default: 'both')
            Change outputs. 'both' returns {'image': image, 'mask': mask}.
        transforms : albumentations.Compose, optional
            An instance of Compose (albumentations pkg) that defines
            augmentation in sequence.
        num_calls : int, optional
            Useful when ``transforms`` is set. Define the total length of the
            dataset. If it is set, it overwrites ``__len__``.
        grayscale : bool (default: False)
            Convert images to grayscale
        grayscale_mode : {'cv2', 'equal', Sequence[float]} (default: 'cv2')
            How to convert to grayscale. If set to 'cv2', it follows opencv
            implementation. Else if set to 'equal', it sums up values along
            channel axis, then divides it by the number of expected channels.
        training : bool (default: True)
            Load training set if True, else load testing one
        idx_invt : sequence of int, optional
            Given indices of images, invert pixel values within their data type.
            e.g. in case of uint8, 0 -> 255, 254 -> 1.

        See Also
        --------
        DSB2018 : Super class
        MaskDataset : Super class of DSB2018
        DatasetInterface : Interface

        """
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            grayscale=grayscale,
            grayscale_mode=grayscale_mode,
            training=training,
            **kwargs
        )
        self.idx_invt = idx_invt

    def get_image(self, p: Path) -> np.ndarray:
        img = Image.open(p)
        img = img.convert(mode='RGB')

        ind = self.file_list.index(p)
        if self.idx_invt is not None:
            if ind in self.idx_invt:
                img = ~np.asarray(img)
                return img
        return np.asarray(img)



"""Dorian's data

It expects below directory structure:
    root_dir
    └── data_folder
        ├── exp0
        │   ├── cond0
        │   └── cond1
        └── exp1
            ├── cond0
            └── cond1

For example:
    Hvo-analyse
    └── data_folder
        ├── H26 GQUAD
        │   ├── Test_Gullermo_2022-12-01
        │   │   ├── FOV1
        │   │   │   └── BG4
        │   │   │       └── fov1-dapi
        │   │   ├── FOV2
        │   │   │   └── BG4
        │   │   │       └── fov2-dapi
        │   │   └── FOV3
        │   │       └── BG4
        │   │           └── fov3-dapi
        │   └── Test_Gu_Untreated
        │       ├── FOV1
        │       │   └── BG4
        │       ├── FOV2
        │       │   └── BG4
        │       └── FOV3
        │           └── BG4
        └── strain_A
            ├── condition_A
            │   ├── fov_A
            │   │   ├── protein_A
            │   │   │   └── ROI1-DAPI
            │   │   └── protein_B
            │   │       └── ROI1-DAPI
            │   └── fov_B
            │       ├── protein_A
            │       │   └── ROI1-DAPI
            │       └── protein_B
            │           └── ROI1-DAPI
            └── condition_B
                ├── fov_A
                │   └── protein_B
                │       └── ROI1-DAPI
                └── fov_B
                    ├── protein_A
                    │   └── ROI1-DAPI
                    └── protein_B
                        └── ROI1-DAPI

"""
import os.path
from functools import cached_property, partial
from pathlib import Path
from typing import List, Optional

import albumentations
import numpy as np
import tifffile

from bioimageloader.base import Dataset


class Dorian(Dataset):
    """Dorian's data

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_calls : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    exp : str, optional
        First parent
    cond : str, optional
        Second parent

    Notes
    -----
    - root_dir is "Hvo-analyse". It will search for "data_folder" in __init__()
    - I named ["H26 GQUAD", "strain_A"] ``exp``, plz find more appropriate name
    - I named ["condition_A", "condition_B"] ``cond``, plz find more appropriate
      name for this one too
    - No FOV selection for now, do you need it?

    Examples
    --------
    Method 3 will be the most useful one for napari integration

    Method 1
    >>> dataset = Dorian('Data/Dorian/Hvo-analyse')  # path to `root_dir`
    >>> print(dataset.exp_list)
    ['H26 GQUAD', 'strain_A']
    >>> print(dataset.cond_list)
    ['H26 GQUAD/Test_Gu_Untreated',
     'H26 GQUAD/Test_Gullermo_2022-12-01',
     'strain_A/condition_A',
     'strain_A/condition_B']
    >>> print(dataset.file_list)
    [PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gu_Untreated/FOV1/BG4/fov1 BG4-dy755-vecta20pc_1_MMStack_Default.ome_DOM_dedrift.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gu_Untreated/FOV1/BG4/fov1-dapi_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gu_Untreated/FOV2/BG4/fov2 BG4-dy755-vecta20pc_1_MMStack_Default.ome_DOM_dedrift.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gu_Untreated/FOV2/BG4/fov2-dapi_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gu_Untreated/FOV3/BG4/fov3 BG4-dy755-vecta20pc_1_MMStack_Default.ome_DOM_dedrift.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gu_Untreated/FOV3/BG4/fov3-dapi_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV1/BG4/fov1-G4-dy755-vecta20pc_1_MMStack_Default.ome_DOM.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV1/BG4/fov1-G4-dy755-vecta20pc_1_MMStack_Default.ome_DOM_dedrift.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV1/BG4/fov1-dapi/fov1-dapi_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV2/BG4/fov2 G4-dy755-vecta20pc_1_MMStack_Default.ome_DOM.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV2/BG4/fov2 G4-dy755-vecta20pc_1_MMStack_Default.ome_DOM_dedrift.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV2/BG4/fov2-dapi/fov2-dapi_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV3/BG4/fov3 G4-dy755-vecta20pc_1_MMStack_Default.ome_DOM.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV3/BG4/fov3 G4-dy755-vecta20pc_1_MMStack_Default.ome_DOM_dedrift.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/H26 GQUAD/Test_Gullermo_2022-12-01/FOV3/BG4/fov3-dapi/fov3-dapi_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_A/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif')]

    Method 2: initialize `exp` or `cond` (static way)
    >>> dataset = Dorian('Data/Dorian/Hvo-analyse', exp='strain_A')
    >>> print(dataset.cond_list)  # dynamic change depending on `exp`
    ['condition_A', 'condition_B']
    >>> print(dataset.file_list)
    [PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_A/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif')]

    Method 3: dynamic `exp` and `cond`
    >>> dataset = Dorian('Data/Dorian/Hvo-analyse')
    >>> dataset.exp = 'strain_A'
    >>> dataset.cond = 'condition_A'
    >>> print(dataset.file_list)
    [PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_A/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_A/fov_B/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif')]
    >>> dataset.cond = 'condition_B'  # change `cond`
    >>> print(dataset.file_list)
    [PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_A/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_A/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_A/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_B/FOV1-DY755 vecta3_1_MMStack_Default.ome.tif'),
     PosixPath('Data/Dorian/Hvo-analyse/data_folder/strain_A/condition_B/fov_B/protein_B/ROI1-DAPI/ROI1-DAPI_MMStack_Default.ome.tif')]

    See Also
    --------
    MaskDataset : Super class
    DatasetInterface : Interface

    """
    # Set acronym
    acronym = 'Dorian'

    def __init__(
        self,
        root_dir: str,
        *,  # only keyword param
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        # specific to this dataset
        cond: Optional[str] = None,
        exp: Optional[str] = None,
        **kwargs
    ):
        self._root_dir = os.path.join(root_dir, 'data_folder')
        self._transforms = transforms
        self._num_calls = num_calls
        # specific to this one here
        self._cond = cond
        self._exp = exp

    @property
    def exp(self):
        return self._exp

    @exp.setter
    def exp(self, val):
        self._exp = val
        if hasattr(self, 'file_list'):
            delattr(self, 'file_list')
        if hasattr(self, 'cond_list'):
            delattr(self, 'cond_list')

    @property
    def cond(self):
        return self._cond

    @cond.setter
    def cond(self, val):
        self._cond = val
        if hasattr(self, 'file_list'):
            delattr(self, 'file_list')

    def get_image(self, p: Path) -> np.ndarray:
        return tifffile.imread(p)

    @staticmethod
    def _fk_contain(p: Path, s: str):
        """filter key"""
        return s in p.as_posix()

    def _filter_file_list_with_keyword(self, file_list: List[Path], keyword: str) -> List[Path]:
        return list(filter(partial(self._fk_contain, s=keyword), self.file_list))

    @cached_property
    def file_list(self) -> List[Path]:
        # Important to decorate with `cached_property` in general
        _file_list = self.root_dir.rglob('*.tif')
        if self.exp is not None:
            _file_list = filter(partial(self._fk_contain, s=self.exp), _file_list)
        if self.cond is not None:
            _file_list = filter(partial(self._fk_contain, s=self.cond), _file_list)
        return sorted(_file_list)

    @cached_property
    def exp_list(self) -> List[str]:
        return [p.name for p in self.root_dir.glob('*')]

    @cached_property
    def cond_list(self) -> List[str]:
        if self.exp is not None:
            subdir = self.root_dir / self.exp
            return [p.relative_to(subdir).as_posix() for p in subdir.glob('*')]
        return [p.relative_to(self.root_dir).as_posix()
                for p in self.root_dir.glob('*/*')]



from pathlib import Path
from typing import Optional, Sequence, Union

import albumentations
import numpy as np
from PIL import Image

from bioimageloader.collections import ComputationalPathology


class ComputationalPathologyNUnet(ComputationalPathology):
    """A Dataset and a Technique for Generalized Nuclear Segmentation for
    Computational Pathology [1]_ (Adopted for NU-Net)

    Notes
    -----
    - Resolution of all images is (1000,1000)
    - gt is converted from annotation recorded in xml format
    - gt has dtype of torch.float64, converted from numpy.uint16, and it has
      value 'num_objects' * 255 because it is base-transformed
    - The origianl dataset provides annotation in xml format, which takes
      long time to parse and to reconstruct mask images dynamically during
      training. Drawing masks beforehand makes training much faster. Use
      `mask_tif` in that case.
    - When `augmenters` is provided, set the `num_calls` argument
      30x1000x1000 -> 16x30=480 patches. Thus, the default `num_calls=720`
      (x1.5)
    - dtype of 'gt' is int16. However, to make batching easier, it will be
      casted to float32
    - Be careful about types of augmenters; avoid interpolation

    References
    ----------
    .. [1] N. Kumar, R. Verma, S. Sharma, S. Bhargava, A. Vahadane, and A.
       Sethi, “A Dataset and a Technique for Generalized Nuclear Segmentation
       for Computational Pathology,” IEEE Transactions on Medical Imaging, vol.
       36, no. 7, pp. 1550–1560, Jul. 2017, doi: 10.1109/TMI.2017.2677499.

    """

    # Dataset's acronym
    acronym = 'ComPath'
    # Hard code resolution to parse annotation (.xml)
    _resolution = (1000, 1000)

    def __init__(
        self,
        # Interface requirement
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional[albumentations.Compose] = None,
        num_calls: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'cv2',
        # Specific to this dataset
        mask_tif: bool = False,
        # NU-Net args
        invt: bool = True,
        **kwargs
    ):
        """
        Parameters
        ----------
        root_dir : str
            Path to root directory
        output : {'image','mask','both'} (default: 'both')
            Change outputs. 'both' returns {'image': image, 'mask': mask}.
        transforms : albumentations.Compose, optional
            An instance of Compose (albumentations pkg) that defines
            augmentation in sequence.
        num_calls : int, optional
            Useful when ``transforms`` is set. Define the total length of the
            dataset. If it is set, it overwrites ``__len__``.
        grayscale : bool (default: False)
            Convert images to grayscale
        grayscale_mode : {'cv2', 'equal', Sequence[float]} (default: 'cv2')
            How to convert to grayscale. If set to 'cv2', it follows opencv
            implementation. Else if set to 'equal', it sums up values along
            channel axis, then divides it by the number of expected channels.
        mask_tif : bool (default: False)
            Instead of parsing every xml file to reconstruct mask image arrays,
            use pre-drawn mask tif files which should reside in the same folder
            as annotation xml files.
        invt : bool (default: True)
            Invert image pixel values

        See Also
        --------
        ComputationalPathology : Super class
        MaskDataset : Super class
        MaskDatasetInterface : Interface of the super class

        """
        super().__init__(
            root_dir=root_dir,
            output=output,
            transforms=transforms,
            num_calls=num_calls,
            grayscale=grayscale,
            grayscale_mode=grayscale_mode,
            mask_tif=mask_tif,
            **kwargs
        )
        self.invt = invt

    def get_image(self, p: Path) -> np.ndarray:
        img = Image.open(p)
        if img.mode == 'RGBA':
            img = img.convert(mode='RGB')
        if self.invt:
            return ~np.asarray(img)
        return np.asarray(img)
